import 'package:async/async.dart';
import 'package:bloc/bloc.dart';
import 'package:bloc_concurrency/bloc_concurrency.dart';
import 'package:dio/dio.dart';
import 'package:meta/meta.dart';

import '../defined.dart';
import '../model/data_paging.dart';
import '../model/exceptions.dart';
import '../utils/logger.dart';
import 'condition_mixins.dart';

part 'events.dart';

part 'states.dart';

/// Base handles normal events
abstract class BaseBloc extends Bloc<BaseEvent, BaseState>
    with NetworkCondition {
  BaseEvent? _currentEvent;
  CancelableOperation? _cancelableOperation;

  BaseBloc() : super(BaseInitState()) {
    reset();
    on<BaseErrorEvent>(
      handleBaseErrorEvent,
      transformer: restartable(),
    );
    on<CancelRunningTaskEvent>(
      handleCancelRunningTaskEvent,
      transformer: restartable(),
    );
    on<ResetBlocEvent>(
      handleResetBlocEvent,
      transformer: restartable(),
    );
  }

  @override
  void onTransition(Transition<BaseEvent, BaseState> transition) {
    super.onTransition(transition);
    if (needLog) {
      logger.d('🚒🚒🚒 TRANSITION BLOC [-$runtimeType-] 🚒🚒🚒\n$transition');
    }
  }

  @override
  Future<void> close() async {
    await _cancelableOperation?.cancel();
    return super.close();
  }

  /// A wrapper of [on] function to ease the handling error flow
  @protected
  void watch<E extends BaseEvent>(
    EventHandler<E, BaseState> handler, {
    EventTransformer<E>? transformer,
  }) {
    on<E>(
      (event, emit) async =>
          await _handleEventWithException(handler, event, emit),
      transformer: transformer,
    );
  }

  /// Throws error state
  @protected
  Future<void> handleBaseErrorEvent(
    BaseErrorEvent event,
    Emitter<BaseState> emit,
  ) async {
    emit(BaseErrorState(
      exception: event.exception,
      tag: event.tag,
    ));
  }

  /// Cancels running tasks
  @protected
  Future<void> handleCancelRunningTaskEvent(
    CancelRunningTaskEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _cancelableOperation?.cancel();
  }

  /// Needs to reset data bloc, includes cancelling all running async tasks
  @protected
  @mustCallSuper
  Future<void> handleResetBlocEvent(
    ResetBlocEvent event,
    Emitter<BaseState> emit,
  ) async {
    await _cancelableOperation?.cancel();
    reset();
  }

  /// Handle exception which is thrown from mapping event-state process
  @protected
  @mustCallSuper
  void handleError(exception, {String? tag}) {
    if (exception is KntException) {
      if (needLog) {
        logger.e('❌❌ KntException: $exception ❌❌');
      }
      add(BaseErrorEvent(
        tag: tag,
        exception: exception,
      ));
      return;
    }
    if (exception is Exception) {
      if (needLog) {
        logger.e('❌❌ Error: ${exception.toString()} ❌❌');
      }
      add(BaseErrorEvent(
        tag: tag,
        exception: KntException(
          KntIssue.undefined,
          additionalInfo: exception.toString(),
        ),
      ));
    }
  }

  /// Reset data
  @protected
  void reset() {}

  /// Whether need to log bloc transition or not
  @protected
  bool get needLog => false;

  /// Wrap cancelable async process with connection status - [needCheckConnection]
  /// and error process if need - [needHandleError]
  @protected
  Future? cancelableFuture(
    Future inner, {
    bool needCheckConnection = true,
    bool needHandleError = true,
  }) async {
    final preProcess =
        needCheckConnection ? () async => await checkConnection() : null;
    final process = processCancelableFuture(inner, preProcess: preProcess);
    return needHandleError
        ? process?.onError(
            (error, stackTrace) => handleError(error, tag: _currentEvent?.tag))
        : process;
  }

  /// Create an async process which can be cancelled
  @protected
  Future? processCancelableFuture(
    Future inner, {
    Future? Function()? preProcess,
  }) async {
    await preProcess?.call();
    await _cancelableOperation?.cancel();
    _cancelableOperation = CancelableOperation.fromFuture(inner);
    return _cancelableOperation?.value;
  }

  Future<void> _handleEventWithException<E extends BaseEvent>(
    EventHandler<E, BaseState> originHandler,
    E event,
    Emitter<BaseState> emit,
  ) async {
    try {
      if (event.needCheckConnection) {
        await checkConnection();
      }
      await originHandler(event, emit);
    } on Exception catch (e) {
      handleError(e, tag: event.tag);
    }
  }
}

typedef CompareFunction<T> = int Function(T, T);

/// Base handles network required events & specific [DioError]
abstract class BaseApiBloc extends BaseBloc {
  @override
  void handleError(exception, {String? tag}) {
    if (exception is DioError) {
      add(BaseErrorEvent(
        tag: tag,
        exception: convertedException(exception),
      ));
      return;
    }
    super.handleError(exception, tag: tag);
  }

  KntException convertedException(DioError exception) {
    return KntException(
      convertedIssue(exception),
      additionalInfo: exception.response?.statusMessage ?? exception.message,
    );
  }

  KntIssue convertedIssue(DioError exception) {
    final httpCode = exception.response?.statusCode;
    return httpCode?.convertedIssue ?? KntIssue.undefined;
  }
}

/// Base handle paging events
abstract class BasePagingBloc<Q, D> extends BaseApiBloc {
  final List<D> _dataList = [];
  late int _currentOffset;
  late bool _loadedAllData;

  BasePagingBloc() : super() {
    watch<BasePagingEvent<Q>>(
      handlePagingEvent,
      transformer: pagingTransformer,
    );
  }

  @override
  Future<void> handleBaseErrorEvent(
    BaseErrorEvent event,
    Emitter<BaseState> emit,
  ) async {
    if (_dataList.isEmpty) {
      // Error occurs at the first query
      return super.handleBaseErrorEvent(event, emit);
    }
    // Otherwise, ignore this error while loading more data
    emit(pagingStateInstance(_dataList, _currentOffset, tag: event.tag));
  }

  @override
  @mustCallSuper
  void reset() {
    _dataList.clear();
    _currentOffset = defaultOffsetPaging;
    _loadedAllData = false;
  }

  @protected
  EventTransformer<BasePagingEvent<Q>> get pagingTransformer => sequential();

  @protected
  Future<void> handlePagingEvent(
    BasePagingEvent<Q> event,
    Emitter<BaseState> emit,
  ) async {
    // Check if all data is loaded
    final isFirstRequest = event.isFirstRequest;
    if (!isFirstRequest && _loadedAllData) {
      return;
    }
    // Check conditions
    if (pagingQueryNeedCheckConnection) {
      await checkConnection();
      // Add other conditions in the future
    }
    if (isFirstRequest) {
      reset();
    }

    // After that, start paging request
    final tag = event.tag;
    emit(BaseLoadingState(tag: tag));
    final responsePaging = await cancelableFuture(
      pagingInProgress(BaseRequestPaging<Q>(
        queryData: event.queryData,
        limit: event.limit,
        offset: _currentOffset,
      )),
      // Already checked connection
      needCheckConnection: false,
      // Need rethrow here to spread out to outer layer
      needHandleError: false,
    );
    if (responsePaging is! BaseResponsePaging<D>) {
      throw KntException(KntIssue.parsing);
    }
    List<D> dataList = responsePaging.dataList;
    if (dataList.length < event.limit) {
      // Reached last page if result has less than limit
      _loadedAllData = true;
    }
    if (dataList.isEmpty && isFirstRequest) {
      // In case at first paging request we have empty data, we could stop
      emit(BaseEmptyResultState(tag: tag));
      return;
    }
    // Update data list and offset
    _dataList.addAll(dataList.where((element) => element != null));
    if (responsePaging.currentOffset == defaultInvalidOffset) {
      _currentOffset += dataList.length;
    } else {
      _currentOffset = responsePaging.currentOffset;
    }
    // Sorting
    if (comparable != null) {
      _dataList.sort(comparable);
    }
    // Show state
    emit(pagingStateInstance(_dataList, _currentOffset, tag: tag));
  }

  /// Handle query process
  @protected
  Future<BaseResponsePaging<D>> pagingInProgress(
      BaseRequestPaging<Q> pagingQuery);

  /// Create instance of paging result
  @protected
  BasePagingResultState<D> pagingStateInstance(
    List<D> dataList,
    int offset, {
    String? tag,
  });

  /// Override this getter when we want to sort out result
  @protected
  CompareFunction<D>? get comparable => null;

  /// Override this getter when we want to check internet connection before querying
  @protected
  bool get pagingQueryNeedCheckConnection => false;
}
